@@ -3,6 +3,54 @@
 module ActiveRecord
   # = Active Record \Relation
   class Relation
+    class ExplainProxy  # :nodoc:
+      def initialize(relation, options)
+        @relation = relation
+        @options  = options
+      end
+
+      def inspect
+        exec_explain { @relation.send(:exec_queries) }
+      end
+
+      def average(column_name)
+        exec_explain { @relation.average(column_name) }
+      end
+
+      def count(column_name = nil)
+        exec_explain { @relation.count(column_name) }
+      end
+
+      def first(limit = nil)
+        exec_explain { @relation.first(limit) }
+      end
+
+      def last(limit = nil)
+        exec_explain { @relation.last(limit) }
+      end
+
+      def maximum(column_name)
+        exec_explain { @relation.maximum(column_name) }
+      end
+
+      def minimum(column_name)
+        exec_explain { @relation.minimum(column_name) }
+      end
+
+      def pluck(*column_names)
+        exec_explain { @relation.pluck(*column_names) }
+      end
+
+      def sum(identity_or_column = nil)
+        exec_explain { @relation.sum(identity_or_column) }
+      end
+
+      private
+        def exec_explain(&block)
+          @relation.exec_explain(@relation.collecting_queries_for_explain { block.call }, @options)
+        end
+    end
+
     MULTI_VALUE_METHODS  = [:includes, :eager_load, :preload, :select, :group,
                             :order, :joins, :left_outer_joins, :references,
                             :extending, :unscope, :optimizer_hints, :annotate,
@@ -12,26 +60,34 @@ module ActiveRecord
                             :reverse_order, :distinct, :create_with, :skip_query_cache]
 
     CLAUSE_METHODS = [:where, :having, :from]
-    INVALID_METHODS_FOR_DELETE_ALL = [:distinct, :with]
+    INVALID_METHODS_FOR_DELETE_ALL = [:distinct, :with, :with_recursive]
 
     VALUE_METHODS = MULTI_VALUE_METHODS + SINGLE_VALUE_METHODS + CLAUSE_METHODS
 
     include Enumerable
     include FinderMethods, Calculations, SpawnMethods, QueryMethods, Batches, Explain, Delegation
+    include SignedId::RelationMethods, TokenFor::RelationMethods
 
-    attr_reader :table, :klass, :loaded, :predicate_builder
+    attr_reader :table, :model, :loaded, :predicate_builder
     attr_accessor :skip_preloading_value
-    alias :model :klass
+    alias :klass :model
     alias :loaded? :loaded
     alias :locked? :lock_value
 
-    def initialize(klass, table: klass.arel_table, predicate_builder: klass.predicate_builder, values: {})
-      @klass  = klass
+    def initialize(model, table: nil, predicate_builder: nil, values: {})
+      if table
+        predicate_builder ||= model.predicate_builder.with(TableMetadata.new(model, table))
+      else
+        table = model.arel_table
+        predicate_builder ||= model.predicate_builder
+      end
+
+      @model  = model
       @table  = table
       @values = values
       @loaded = false
       @predicate_builder = predicate_builder
-      @delegate_to_klass = false
+      @delegate_to_model = false
       @future_result = nil
       @records = nil
       @async = false
@@ -44,7 +100,7 @@ module ActiveRecord
     end
 
     def bind_attribute(name, value) # :nodoc:
-      if reflection = klass._reflect_on_association(name)
+      if reflection = model._reflect_on_association(name)
         name = reflection.foreign_key
         value = value.read_attribute(reflection.association_primary_key) unless value.nil?
       end
@@ -207,18 +263,22 @@ module ActiveRecord
     #   the problem of running out of integers, if the underlying table is still stuck on a primary
     #   key of type int (note: All \Rails apps since 5.1+ have defaulted to bigint, which is not liable
     #   to this problem).
+    # * Columns with unique database constraints should not have uniqueness validations defined,
+    #   otherwise #create will fail due to validation errors and #find_by will never be called.
     #
     # This method will return a record if all given attributes are covered by unique constraints
     # (unless the INSERT -> DELETE -> SELECT race condition is triggered), but if creation was attempted
     # and failed due to validation errors it won't be persisted, you get what #create returns in
     # such situation.
     def create_or_find_by(attributes, &block)
-      transaction(requires_new: true) { create(attributes, &block) }
-    rescue ActiveRecord::RecordNotUnique
-      if connection.transaction_open?
-        where(attributes).lock.find_by!(attributes)
-      else
-        find_by!(attributes)
+      with_connection do |connection|
+        transaction(requires_new: true) { create(attributes, &block) }
+      rescue ActiveRecord::RecordNotUnique
+        if connection.transaction_open?
+          where(attributes).lock.find_by!(attributes)
+        else
+          find_by!(attributes)
+        end
       end
     end
 
@@ -226,12 +286,14 @@ module ActiveRecord
     # {create!}[rdoc-ref:Persistence::ClassMethods#create!] so an exception
     # is raised if the created record is invalid.
     def create_or_find_by!(attributes, &block)
-      transaction(requires_new: true) { create!(attributes, &block) }
-    rescue ActiveRecord::RecordNotUnique
-      if connection.transaction_open?
-        where(attributes).lock.find_by!(attributes)
-      else
-        find_by!(attributes)
+      with_connection do |connection|
+        transaction(requires_new: true) { create!(attributes, &block) }
+      rescue ActiveRecord::RecordNotUnique
+        if connection.transaction_open?
+          where(attributes).lock.find_by!(attributes)
+        else
+          find_by!(attributes)
+        end
       end
     end
 
@@ -245,13 +307,30 @@ module ActiveRecord
     # returns the result as a string. The string is formatted imitating the
     # ones printed by the database shell.
     #
+    #   User.all.explain
+    #   # EXPLAIN SELECT `users`.* FROM `users`
+    #   # ...
+    #
     # Note that this method actually runs the queries, since the results of some
     # are needed by the next ones when eager loading is going on.
     #
+    # To run EXPLAIN on queries created by +first+, +pluck+ and +count+, call
+    # these methods on +explain+:
+    #
+    #   User.all.explain.count
+    #   # EXPLAIN SELECT COUNT(*) FROM `users`
+    #   # ...
+    #
+    # The column name can be passed if required:
+    #
+    #   User.all.explain.maximum(:id)
+    #   # EXPLAIN SELECT MAX(`users`.`id`) FROM `users`
+    #   # ...
+    #
     # Please see further details in the
     # {Active Record Query Interface guide}[https://guides.rubyonrails.org/active_record_querying.html#running-explain].
     def explain(*options)
-      exec_explain(collecting_queries_for_explain { exec_queries }, options)
+      ExplainProxy.new(self, options)
     end
 
     # Converts relation objects to Array.
@@ -358,14 +437,14 @@ module ActiveRecord
     #   Product.where("name like ?", "%Game%").cache_key(:last_reviewed_at)
     def cache_key(timestamp_column = "updated_at")
       @cache_keys ||= {}
-      @cache_keys[timestamp_column] ||= klass.collection_cache_key(self, timestamp_column)
+      @cache_keys[timestamp_column] ||= model.collection_cache_key(self, timestamp_column)
     end
 
     def compute_cache_key(timestamp_column = :updated_at) # :nodoc:
       query_signature = ActiveSupport::Digest.hexdigest(to_sql)
-      key = "#{klass.model_name.cache_key}/query-#{query_signature}"
+      key = "#{model.model_name.cache_key}/query-#{query_signature}"
 
-      if collection_cache_versioning
+      if model.collection_cache_versioning
         key
       else
         "#{key}-#{compute_cache_version(timestamp_column)}"
@@ -384,7 +463,7 @@ module ActiveRecord
     #
     #    SELECT COUNT(*), MAX("products"."updated_at") FROM "products" WHERE (name like '%Cosmic Encounter%')
     def cache_version(timestamp_column = :updated_at)
-      if collection_cache_versioning
+      if model.collection_cache_versioning
         @cache_versions ||= {}
         @cache_versions[timestamp_column] ||= compute_cache_version(timestamp_column)
       end
@@ -401,33 +480,35 @@ module ActiveRecord
       else
         collection = eager_loading? ? apply_join_dependency : self
 
-        column = connection.visitor.compile(table[timestamp_column])
-        select_values = "COUNT(*) AS #{connection.quote_column_name("size")}, MAX(%s) AS timestamp"
+        with_connection do |c|
+          column = c.visitor.compile(table[timestamp_column])
+          select_values = "COUNT(*) AS #{model.adapter_class.quote_column_name("size")}, MAX(%s) AS timestamp"
 
-        if collection.has_limit_or_offset?
-          query = collection.select("#{column} AS collection_cache_key_timestamp")
-          query._select!(table[Arel.star]) if distinct_value && collection.select_values.empty?
-          subquery_alias = "subquery_for_cache_key"
-          subquery_column = "#{subquery_alias}.collection_cache_key_timestamp"
-          arel = query.build_subquery(subquery_alias, select_values % subquery_column)
-        else
-          query = collection.unscope(:order)
-          query.select_values = [select_values % column]
-          arel = query.arel
-        end
+          if collection.has_limit_or_offset?
+            query = collection.select("#{column} AS collection_cache_key_timestamp")
+            query._select!(table[Arel.star]) if distinct_value && collection.select_values.empty?
+            subquery_alias = "subquery_for_cache_key"
+            subquery_column = "#{subquery_alias}.collection_cache_key_timestamp"
+            arel = query.build_subquery(subquery_alias, select_values % subquery_column)
+          else
+            query = collection.unscope(:order)
+            query.select_values = [select_values % column]
+            arel = query.arel
+          end
 
-        size, timestamp = connection.select_rows(arel, nil).first
+          size, timestamp = c.select_rows(arel, nil).first
 
-        if size
-          column_type = klass.type_for_attribute(timestamp_column)
-          timestamp = column_type.deserialize(timestamp)
-        else
-          size = 0
+          if size
+            column_type = model.type_for_attribute(timestamp_column)
+            timestamp = column_type.deserialize(timestamp)
+          else
+            size = 0
+          end
         end
       end
 
       if timestamp
-        "#{size}-#{timestamp.utc.to_fs(cache_timestamp_format)}"
+        "#{size}-#{timestamp.utc.to_fs(model.cache_timestamp_format)}"
       else
         "#{size}"
       end
@@ -458,7 +539,7 @@ module ActiveRecord
     # Please check unscoped if you want to remove all previous scopes (including
     # the default_scope) during the execution of a block.
     def scoping(all_queries: nil, &block)
-      registry = klass.scope_registry
+      registry = model.scope_registry
       if global_scope?(registry) && all_queries == false
         raise ArgumentError, "Scoping is set to apply to all queries and cannot be unset in a nested block."
       elsif already_in_scope?(registry)
@@ -469,11 +550,11 @@ module ActiveRecord
     end
 
     def _exec_scope(...) # :nodoc:
-      @delegate_to_klass = true
-      registry = klass.scope_registry
+      @delegate_to_model = true
+      registry = model.scope_registry
       _scoping(nil, registry) { instance_exec(...) || self }
     ensure
-      @delegate_to_klass = false
+      @delegate_to_model = false
     end
 
     # Updates all records in the current relation with details given. This method constructs a single SQL UPDATE
@@ -510,31 +591,38 @@ module ActiveRecord
       return 0 if @none
 
       if updates.is_a?(Hash)
-        if klass.locking_enabled? &&
-            !updates.key?(klass.locking_column) &&
-            !updates.key?(klass.locking_column.to_sym)
-          attr = table[klass.locking_column]
+        if model.locking_enabled? &&
+            !updates.key?(model.locking_column) &&
+            !updates.key?(model.locking_column.to_sym)
+          attr = table[model.locking_column]
           updates[attr.name] = _increment_attribute(attr)
         end
         values = _substitute_values(updates)
       else
-        values = Arel.sql(klass.sanitize_sql_for_assignment(updates, table.name))
+        values = Arel.sql(model.sanitize_sql_for_assignment(updates, table.name))
       end
 
-      arel = eager_loading? ? apply_join_dependency.arel : build_arel
-      arel.source.left = table
+      model.with_connection do |c|
+        arel = eager_loading? ? apply_join_dependency.arel : build_arel(c)
+        arel.source.left = table
 
-      group_values_arel_columns = arel_columns(group_values.uniq)
-      having_clause_ast = having_clause.ast unless having_clause.empty?
-      stmt = arel.compile_update(values, table[primary_key], having_clause_ast, group_values_arel_columns)
-      klass.connection.update(stmt, "#{klass} Update All").tap { reset }
+        group_values_arel_columns = arel_columns(group_values.uniq)
+        having_clause_ast = having_clause.ast unless having_clause.empty?
+        key = if model.composite_primary_key?
+          primary_key.map { |pk| table[pk] }
+        else
+          table[primary_key]
+        end
+        stmt = arel.compile_update(values, key, having_clause_ast, group_values_arel_columns)
+        c.update(stmt, "#{model} Update All").tap { reset }
+      end
     end
 
     def update(id = :all, attributes) # :nodoc:
       if id == :all
         each { |record| record.update(attributes) }
       else
-        klass.update(id, attributes)
+        model.update(id, attributes)
       end
     end
 
@@ -542,10 +630,287 @@ module ActiveRecord
       if id == :all
         each { |record| record.update!(attributes) }
       else
-        klass.update!(id, attributes)
+        model.update!(id, attributes)
       end
     end
 
+
+    # Inserts a single record into the database in a single SQL INSERT
+    # statement. It does not instantiate any models nor does it trigger
+    # Active Record callbacks or validations. Though passed values
+    # go through Active Record's type casting and serialization.
+    #
+    # See #insert_all for documentation.
+    def insert(attributes, returning: nil, unique_by: nil, record_timestamps: nil)
+      insert_all([ attributes ], returning: returning, unique_by: unique_by, record_timestamps: record_timestamps)
+    end
+
+    # Inserts multiple records into the database in a single SQL INSERT
+    # statement. It does not instantiate any models nor does it trigger
+    # Active Record callbacks or validations. Though passed values
+    # go through Active Record's type casting and serialization.
+    #
+    # The +attributes+ parameter is an Array of Hashes. Every Hash determines
+    # the attributes for a single row and must have the same keys.
+    #
+    # Rows are considered to be unique by every unique index on the table. Any
+    # duplicate rows are skipped.
+    # Override with <tt>:unique_by</tt> (see below).
+    #
+    # Returns an ActiveRecord::Result with its contents based on
+    # <tt>:returning</tt> (see below).
+    #
+    # ==== Options
+    #
+    # [:returning]
+    #   (PostgreSQL, SQLite3, and MariaDB only) An array of attributes to return for all successfully
+    #   inserted records, which by default is the primary key.
+    #   Pass <tt>returning: %w[ id name ]</tt> for both id and name
+    #   or <tt>returning: false</tt> to omit the underlying <tt>RETURNING</tt> SQL
+    #   clause entirely.
+    #
+    #   You can also pass an SQL string if you need more control on the return values
+    #   (for example, <tt>returning: Arel.sql("id, name as new_name")</tt>).
+    #
+    # [:unique_by]
+    #   (PostgreSQL and SQLite only) By default rows are considered to be unique
+    #   by every unique index on the table. Any duplicate rows are skipped.
+    #
+    #   To skip rows according to just one unique index pass <tt>:unique_by</tt>.
+    #
+    #   Consider a Book model where no duplicate ISBNs make sense, but if any
+    #   row has an existing id, or is not unique by another unique index,
+    #   ActiveRecord::RecordNotUnique is raised.
+    #
+    #   Unique indexes can be identified by columns or name:
+    #
+    #     unique_by: :isbn
+    #     unique_by: %i[ author_id name ]
+    #     unique_by: :index_books_on_isbn
+    #
+    # [:record_timestamps]
+    #   By default, automatic setting of timestamp columns is controlled by
+    #   the model's <tt>record_timestamps</tt> config, matching typical
+    #   behavior.
+    #
+    #   To override this and force automatic setting of timestamp columns one
+    #   way or the other, pass <tt>:record_timestamps</tt>:
+    #
+    #     record_timestamps: true  # Always set timestamps automatically
+    #     record_timestamps: false # Never set timestamps automatically
+    #
+    # Because it relies on the index information from the database
+    # <tt>:unique_by</tt> is recommended to be paired with
+    # Active Record's schema_cache.
+    #
+    # ==== Example
+    #
+    #   # Insert records and skip inserting any duplicates.
+    #   # Here "Eloquent Ruby" is skipped because its id is not unique.
+    #
+    #   Book.insert_all([
+    #     { id: 1, title: "Rework", author: "David" },
+    #     { id: 1, title: "Eloquent Ruby", author: "Russ" }
+    #   ])
+    #
+    #   # insert_all works on chained scopes, and you can use create_with
+    #   # to set default attributes for all inserted records.
+    #
+    #   author.books.create_with(created_at: Time.now).insert_all([
+    #     { id: 1, title: "Rework" },
+    #     { id: 2, title: "Eloquent Ruby" }
+    #   ])
+    def insert_all(attributes, returning: nil, unique_by: nil, record_timestamps: nil)
+      InsertAll.execute(self, attributes, on_duplicate: :skip, returning: returning, unique_by: unique_by, record_timestamps: record_timestamps)
+    end
+
+    # Inserts a single record into the database in a single SQL INSERT
+    # statement. It does not instantiate any models nor does it trigger
+    # Active Record callbacks or validations. Though passed values
+    # go through Active Record's type casting and serialization.
+    #
+    # See #insert_all! for more.
+    def insert!(attributes, returning: nil, record_timestamps: nil)
+      insert_all!([ attributes ], returning: returning, record_timestamps: record_timestamps)
+    end
+
+    # Inserts multiple records into the database in a single SQL INSERT
+    # statement. It does not instantiate any models nor does it trigger
+    # Active Record callbacks or validations. Though passed values
+    # go through Active Record's type casting and serialization.
+    #
+    # The +attributes+ parameter is an Array of Hashes. Every Hash determines
+    # the attributes for a single row and must have the same keys.
+    #
+    # Raises ActiveRecord::RecordNotUnique if any rows violate a
+    # unique index on the table. In that case, no rows are inserted.
+    #
+    # To skip duplicate rows, see #insert_all. To replace them, see #upsert_all.
+    #
+    # Returns an ActiveRecord::Result with its contents based on
+    # <tt>:returning</tt> (see below).
+    #
+    # ==== Options
+    #
+    # [:returning]
+    #   (PostgreSQL, SQLite3, and MariaDB only) An array of attributes to return for all successfully
+    #   inserted records, which by default is the primary key.
+    #   Pass <tt>returning: %w[ id name ]</tt> for both id and name
+    #   or <tt>returning: false</tt> to omit the underlying <tt>RETURNING</tt> SQL
+    #   clause entirely.
+    #
+    #   You can also pass an SQL string if you need more control on the return values
+    #   (for example, <tt>returning: Arel.sql("id, name as new_name")</tt>).
+    #
+    # [:record_timestamps]
+    #   By default, automatic setting of timestamp columns is controlled by
+    #   the model's <tt>record_timestamps</tt> config, matching typical
+    #   behavior.
+    #
+    #   To override this and force automatic setting of timestamp columns one
+    #   way or the other, pass <tt>:record_timestamps</tt>:
+    #
+    #     record_timestamps: true  # Always set timestamps automatically
+    #     record_timestamps: false # Never set timestamps automatically
+    #
+    # ==== Examples
+    #
+    #   # Insert multiple records
+    #   Book.insert_all!([
+    #     { title: "Rework", author: "David" },
+    #     { title: "Eloquent Ruby", author: "Russ" }
+    #   ])
+    #
+    #   # Raises ActiveRecord::RecordNotUnique because "Eloquent Ruby"
+    #   # does not have a unique id.
+    #   Book.insert_all!([
+    #     { id: 1, title: "Rework", author: "David" },
+    #     { id: 1, title: "Eloquent Ruby", author: "Russ" }
+    #   ])
+    def insert_all!(attributes, returning: nil, record_timestamps: nil)
+      InsertAll.execute(self, attributes, on_duplicate: :raise, returning: returning, record_timestamps: record_timestamps)
+    end
+
+    # Updates or inserts (upserts) a single record into the database in a
+    # single SQL INSERT statement. It does not instantiate any models nor does
+    # it trigger Active Record callbacks or validations. Though passed values
+    # go through Active Record's type casting and serialization.
+    #
+    # See #upsert_all for documentation.
+    def upsert(attributes, **kwargs)
+      upsert_all([ attributes ], **kwargs)
+    end
+
+    # Updates or inserts (upserts) multiple records into the database in a
+    # single SQL INSERT statement. It does not instantiate any models nor does
+    # it trigger Active Record callbacks or validations. Though passed values
+    # go through Active Record's type casting and serialization.
+    #
+    # The +attributes+ parameter is an Array of Hashes. Every Hash determines
+    # the attributes for a single row and must have the same keys.
+    #
+    # Returns an ActiveRecord::Result with its contents based on
+    # <tt>:returning</tt> (see below).
+    #
+    # By default, +upsert_all+ will update all the columns that can be updated when
+    # there is a conflict. These are all the columns except primary keys, read-only
+    # columns, and columns covered by the optional +unique_by+.
+    #
+    # ==== Options
+    #
+    # [:returning]
+    #   (PostgreSQL, SQLite3, and MariaDB only) An array of attributes to return for all successfully
+    #   upserted records, which by default is the primary key.
+    #   Pass <tt>returning: %w[ id name ]</tt> for both id and name
+    #   or <tt>returning: false</tt> to omit the underlying <tt>RETURNING</tt> SQL
+    #   clause entirely.
+    #
+    #   You can also pass an SQL string if you need more control on the return values
+    #   (for example, <tt>returning: Arel.sql("id, name as new_name")</tt>).
+    #
+    # [:unique_by]
+    #   (PostgreSQL and SQLite only) By default rows are considered to be unique
+    #   by every unique index on the table. Any duplicate rows are skipped.
+    #
+    #   To skip rows according to just one unique index pass <tt>:unique_by</tt>.
+    #
+    #   Consider a Book model where no duplicate ISBNs make sense, but if any
+    #   row has an existing id, or is not unique by another unique index,
+    #   ActiveRecord::RecordNotUnique is raised.
+    #
+    #   Unique indexes can be identified by columns or name:
+    #
+    #     unique_by: :isbn
+    #     unique_by: %i[ author_id name ]
+    #     unique_by: :index_books_on_isbn
+    #
+    # Because it relies on the index information from the database
+    # <tt>:unique_by</tt> is recommended to be paired with
+    # Active Record's schema_cache.
+    #
+    # [:on_duplicate]
+    #   Configure the SQL update sentence that will be used in case of conflict.
+    #
+    #   NOTE: If you use this option you must provide all the columns you want to update
+    #   by yourself.
+    #
+    #   Example:
+    #
+    #     Commodity.upsert_all(
+    #       [
+    #         { id: 2, name: "Copper", price: 4.84 },
+    #         { id: 4, name: "Gold", price: 1380.87 },
+    #         { id: 6, name: "Aluminium", price: 0.35 }
+    #       ],
+    #       on_duplicate: Arel.sql("price = GREATEST(commodities.price, EXCLUDED.price)")
+    #     )
+    #
+    #   See the related +:update_only+ option. Both options can't be used at the same time.
+    #
+    # [:update_only]
+    #   Provide a list of column names that will be updated in case of conflict. If not provided,
+    #   +upsert_all+ will update all the columns that can be updated. These are all the columns
+    #   except primary keys, read-only columns, and columns covered by the optional +unique_by+
+    #
+    #   Example:
+    #
+    #     Commodity.upsert_all(
+    #       [
+    #         { id: 2, name: "Copper", price: 4.84 },
+    #         { id: 4, name: "Gold", price: 1380.87 },
+    #         { id: 6, name: "Aluminium", price: 0.35 }
+    #       ],
+    #       update_only: [:price] # Only prices will be updated
+    #     )
+    #
+    #   See the related +:on_duplicate+ option. Both options can't be used at the same time.
+    #
+    # [:record_timestamps]
+    #   By default, automatic setting of timestamp columns is controlled by
+    #   the model's <tt>record_timestamps</tt> config, matching typical
+    #   behavior.
+    #
+    #   To override this and force automatic setting of timestamp columns one
+    #   way or the other, pass <tt>:record_timestamps</tt>:
+    #
+    #     record_timestamps: true  # Always set timestamps automatically
+    #     record_timestamps: false # Never set timestamps automatically
+    #
+    # ==== Examples
+    #
+    #   # Inserts multiple records, performing an upsert when records have duplicate ISBNs.
+    #   # Here "Eloquent Ruby" overwrites "Rework" because its ISBN is duplicate.
+    #
+    #   Book.upsert_all([
+    #     { title: "Rework", author: "David", isbn: "1" },
+    #     { title: "Eloquent Ruby", author: "Russ", isbn: "1" }
+    #   ], unique_by: :isbn)
+    #
+    #   Book.find_by(isbn: "1").title # => "Eloquent Ruby"
+    def upsert_all(attributes, on_duplicate: :update, update_only: nil, returning: nil, unique_by: nil, record_timestamps: nil)
+      InsertAll.execute(self, attributes, on_duplicate: on_duplicate, update_only: update_only, returning: returning, unique_by: unique_by, record_timestamps: record_timestamps)
+    end
+
     # Updates the counters of the records in the current relation.
     #
     # ==== Parameters
@@ -571,7 +936,7 @@ module ActiveRecord
         names = touch if touch != true
         names = Array.wrap(names)
         options = names.extract_options!
-        touch_updates = klass.touch_attributes_with_time(*names, **options)
+        touch_updates = model.touch_attributes_with_time(*names, **options)
         updates.merge!(touch_updates) unless touch_updates.empty?
       end
 
@@ -602,7 +967,7 @@ module ActiveRecord
     #   Person.where(name: 'David').touch_all
     #   # => "UPDATE \"people\" SET \"updated_at\" = '2018-01-04 22:55:23.132670' WHERE \"people\".\"name\" = 'David'"
     def touch_all(*names, time: nil)
-      update_all klass.touch_attributes_with_time(*names, time: time)
+      update_all model.touch_attributes_with_time(*names, time: time)
     end
 
     # Destroys the records by instantiating each
@@ -654,14 +1019,79 @@ module ActiveRecord
         raise ActiveRecordError.new("delete_all doesn't support #{invalid_methods.join(', ')}")
       end
 
-      arel = eager_loading? ? apply_join_dependency.arel : build_arel
-      arel.source.left = table
+      model.with_connection do |c|
+        arel = eager_loading? ? apply_join_dependency.arel : build_arel(c)
+        arel.source.left = table
 
-      group_values_arel_columns = arel_columns(group_values.uniq)
-      having_clause_ast = having_clause.ast unless having_clause.empty?
-      stmt = arel.compile_delete(table[primary_key], having_clause_ast, group_values_arel_columns)
+        group_values_arel_columns = arel_columns(group_values.uniq)
+        having_clause_ast = having_clause.ast unless having_clause.empty?
+        key = if model.composite_primary_key?
+          primary_key.map { |pk| table[pk] }
+        else
+          table[primary_key]
+        end
+        stmt = arel.compile_delete(key, having_clause_ast, group_values_arel_columns)
 
-      klass.connection.delete(stmt, "#{klass} Delete All").tap { reset }
+        c.delete(stmt, "#{model} Delete All").tap { reset }
+      end
+    end
+
+    # Deletes the row with a primary key matching the +id+ argument, using an
+    # SQL +DELETE+ statement, and returns the number of rows deleted. Active
+    # Record objects are not instantiated, so the object's callbacks are not
+    # executed, including any <tt>:dependent</tt> association options.
+    #
+    # You can delete multiple rows at once by passing an Array of <tt>id</tt>s.
+    #
+    # Note: Although it is often much faster than the alternative, #destroy,
+    # skipping callbacks might bypass business logic in your application
+    # that ensures referential integrity or performs other essential jobs.
+    #
+    # ==== Examples
+    #
+    #   # Delete a single row
+    #   Todo.delete(1)
+    #
+    #   # Delete multiple rows
+    #   Todo.delete([2,3,4])
+    def delete(id_or_array)
+      return 0 if id_or_array.nil? || (id_or_array.is_a?(Array) && id_or_array.empty?)
+
+      where(model.primary_key => id_or_array).delete_all
+    end
+
+
+    # Destroy an object (or multiple objects) that has the given id. The object is instantiated first,
+    # therefore all callbacks and filters are fired off before the object is deleted. This method is
+    # less efficient than #delete but allows cleanup methods and other actions to be run.
+    #
+    # This essentially finds the object (or multiple objects) with the given id, creates a new object
+    # from the attributes, and then calls destroy on it.
+    #
+    # ==== Parameters
+    #
+    # * +id+ - This should be the id or an array of ids to be destroyed.
+    #
+    # ==== Examples
+    #
+    #   # Destroy a single object
+    #   Todo.destroy(1)
+    #
+    #   # Destroy multiple objects
+    #   todos = [1,2,3]
+    #   Todo.destroy(todos)
+    def destroy(id)
+      multiple_ids = if model.composite_primary_key?
+        id.first.is_a?(Array)
+      else
+        id.is_a?(Array)
+      end
+
+      if multiple_ids
+        find(id).each(&:destroy)
+      else
+        find(id).destroy
+      end
     end
 
     # Finds and destroys all records matching the specified conditions.
@@ -701,30 +1131,39 @@ module ActiveRecord
     # for queries to actually be executed concurrently. Otherwise it defaults to
     # executing them in the foreground.
     #
-    # +load_async+ will also fall back to executing in the foreground in the test environment when transactional
-    # fixtures are enabled.
-    #
     # If the query was actually executed in the background, the Active Record logs will show
     # it by prefixing the log line with <tt>ASYNC</tt>:
     #
     #   ASYNC Post Load (0.0ms) (db time 2ms)  SELECT "posts".* FROM "posts" LIMIT 100
     def load_async
-      return load if !connection.async_enabled?
+      with_connection do |c|
+        return load if !c.async_enabled?
 
-      unless loaded?
-        result = exec_main_query(async: connection.current_transaction.closed?)
+        unless loaded?
+          result = exec_main_query(async: !c.current_transaction.joinable?)
 
-        if result.is_a?(Array)
-          @records = result
-        else
-          @future_result = result
+          if result.is_a?(Array)
+            @records = result
+          else
+            @future_result = result
+          end
+          @loaded = true
         end
-        @loaded = true
       end
 
       self
     end
 
+    def then(&block) # :nodoc:
+      if @future_result
+        @future_result.then do
+          yield self
+        end
+      else
+        super
+      end
+    end
+
     # Returns <tt>true</tt> if the relation was scheduled on the background
     # thread pool.
     def scheduled?
@@ -755,7 +1194,7 @@ module ActiveRecord
     def reset
       @future_result&.cancel
       @future_result = nil
-      @delegate_to_klass = false
+      @delegate_to_model = false
       @to_sql = @arel = @loaded = @should_eager_load = nil
       @offsets = @take = nil
       @cache_keys = nil
@@ -775,8 +1214,9 @@ module ActiveRecord
           relation.to_sql
         end
       else
-        conn = klass.connection
-        conn.unprepared_statement { conn.to_sql(arel) }
+        model.with_connection do |conn|
+          conn.unprepared_statement { conn.to_sql(arel) }
+        end
       end
     end
 
@@ -784,12 +1224,12 @@ module ActiveRecord
     #
     #   User.where(name: 'Oscar').where_values_hash
     #   # => {name: "Oscar"}
-    def where_values_hash(relation_table_name = klass.table_name) # :nodoc:
+    def where_values_hash(relation_table_name = model.table_name) # :nodoc:
       where_clause.to_h(relation_table_name)
     end
 
     def scope_for_create
-      hash = where_clause.to_h(klass.table_name, equality_only: true)
+      hash = where_clause.to_h(model.table_name, equality_only: true)
       create_with_value.each { |k, v| hash[k.to_s] = v } unless create_with_value.empty?
       hash
     end
@@ -835,6 +1275,10 @@ module ActiveRecord
       records.blank?
     end
 
+    def readonly?
+      readonly_value
+    end
+
     def values
       @values.dup
     end
@@ -853,7 +1297,7 @@ module ActiveRecord
     end
 
     def empty_scope? # :nodoc:
-      @values == klass.unscoped.values
+      @values == model.unscoped.values
     end
 
     def has_limit_or_offset? # :nodoc:
@@ -861,7 +1305,7 @@ module ActiveRecord
     end
 
     def alias_tracker(joins = [], aliases = nil) # :nodoc:
-      ActiveRecord::Associations::AliasTracker.create(connection, table.name, joins, aliases)
+      ActiveRecord::Associations::AliasTracker.create(model.connection_pool, table.name, joins, aliases)
     end
 
     class StrictLoadingScope # :nodoc:
@@ -891,54 +1335,58 @@ module ActiveRecord
 
     private
       def already_in_scope?(registry)
-        @delegate_to_klass && registry.current_scope(klass, true)
+        @delegate_to_model && registry.current_scope(model, true)
       end
 
       def global_scope?(registry)
-        registry.global_current_scope(klass, true)
+        registry.global_current_scope(model, true)
       end
 
       def current_scope_restoring_block(&block)
-        current_scope = klass.current_scope(true)
+        current_scope = model.current_scope(true)
         -> record do
-          klass.current_scope = current_scope
+          model.current_scope = current_scope
           yield record if block_given?
         end
       end
 
       def _new(attributes, &block)
-        klass.new(attributes, &block)
+        model.new(attributes, &block)
       end
 
       def _create(attributes, &block)
-        klass.create(attributes, &block)
+        model.create(attributes, &block)
       end
 
       def _create!(attributes, &block)
-        klass.create!(attributes, &block)
+        model.create!(attributes, &block)
       end
 
       def _scoping(scope, registry, all_queries = false)
-        previous = registry.current_scope(klass, true)
-        registry.set_current_scope(klass, scope)
+        previous = registry.current_scope(model, true)
+        registry.set_current_scope(model, scope)
 
         if all_queries
-          previous_global = registry.global_current_scope(klass, true)
-          registry.set_global_current_scope(klass, scope)
+          previous_global = registry.global_current_scope(model, true)
+          registry.set_global_current_scope(model, scope)
         end
         yield
       ensure
-        registry.set_current_scope(klass, previous)
+        registry.set_current_scope(model, previous)
         if all_queries
-          registry.set_global_current_scope(klass, previous_global)
+          registry.set_global_current_scope(model, previous_global)
         end
       end
 
       def _substitute_values(values)
         values.map do |name, value|
           attr = table[name]
-          unless Arel.arel_node?(value)
-            type = klass.type_for_attribute(attr.name)
+          if Arel.arel_node?(value)
+            if value.is_a?(Arel::Nodes::SqlLiteral)
+              value = Arel::Nodes::Grouping.new(value)
+            end
+          else
+            type = model.type_for_attribute(attr.name)
             value = predicate_builder.build_bind_attribute(attr.name, type.cast(value))
           end
           [attr, value]
@@ -975,7 +1423,7 @@ module ActiveRecord
       def exec_main_query(async: false)
         if @none
           if async
-            return FutureResult::Complete.new([])
+            return FutureResult.wrap([])
           else
             return []
           end
@@ -985,17 +1433,21 @@ module ActiveRecord
           if where_clause.contradiction?
             [].freeze
           elsif eager_loading?
-            apply_join_dependency do |relation, join_dependency|
-              if relation.null_relation?
-                [].freeze
-              else
-                relation = join_dependency.apply_column_aliases(relation)
-                @_join_dependency = join_dependency
-                connection.select_all(relation.arel, "SQL", async: async)
+            model.with_connection do |c|
+              apply_join_dependency do |relation, join_dependency|
+                if relation.null_relation?
+                  [].freeze
+                else
+                  relation = join_dependency.apply_column_aliases(relation)
+                  @_join_dependency = join_dependency
+                  c.select_all(relation.arel, "SQL", async: async)
+                end
               end
             end
           else
-            klass._query_by_sql(arel, async: async)
+            model.with_connection do |c|
+              model._query_by_sql(c, arel, async: async)
+            end
           end
         end
       end
@@ -1007,13 +1459,13 @@ module ActiveRecord
           @_join_dependency = nil
           records
         else
-          klass._load_from_sql(rows, &block).freeze
+          model._load_from_sql(rows, &block).freeze
         end
       end
 
       def skip_query_cache_if_necessary(&block)
         if skip_query_cache_value
-          uncached(&block)
+          model.uncached(&block)
         else
           yield
         end
